# coding: utf-8

import multiprocessing
import threading
import collections

import pronto.utils


class Relationship(object):
    """
    A Relationship object.

    The Relationship class actually behaves as a factory, creating new
    relationships via the default Python syntax only if no relationship
    of the same name are present in the class py:attribute:: _instances
    (a dictionnary containing memoized relationships).


    .. note::

       Relationships are pickable and always refer to the same adress even
       after being pickled and unpickled, but that requires to use at least
       pickle protocol 2 (which is not default on Python 2, so take care !)

       .. example::

          >>> import pronto
          >>> import io, pickle
          >>>
          >>> src = io.BytesIO()
          >>> p = pickle.Pickler(src, pickle.HIGHEST_PROTOCOL)
          >>>
          >>> isa = pronto.Relationship('is_a')
          >>> isa_id = id(isa)
          >>>
          >>> p.dump(isa)
          >>> dst = io.BytesIO(src.getvalue())
          >>>
          >>> u = pickle.Unpickler(dst)
          >>> new_isa = u.load()
          >>>
          >>> id(new_isa) == isa_id
          True
          >>> # what's that black magic ?!

    """

    _instances = collections.OrderedDict()
    _lock = multiprocessing.Lock()
    _tlock = threading.Lock()

    def __init__(self, obo_name, symmetry=None, transitivity=None,
                 reflexivity=None, complementary=None, prefix=None,
                 direction=None, comment=None, aliases=None):
        """Instantiate a new relationship.

        Parameters:
            obo_name (str): the name of the relationship as it appears
                in obo files (such as is_a, has_part, etc.)
            symetry (bool or None): the symetry of the relationship
            transitivity (bool or None): the transitivity of the relationship.
            reflexivity (bool or None): the reflexivity of the relationship.
            complementary (string or None): if any, the obo_name of the
                complementary relationship.
            direction (string or None): if any, the direction of the
                relationship (can be 'topdown', 'bottomup', 'horizontal').
                A relationship with a direction set as 'topdown' will be
                counted as _childhooding_ when using the Term.children
                property.
            comment (string or None): comments about the Relationship
            aliases (list or None): a list of names that are synonyms to
                this Relationship obo_name

        .. note::
            For :symetry, transitivity, reflexivity, the allowed values are
            the following:
                - True for reflexive, transitive, symmetric
                - False for areflexive, atransitive, asymmetric
                - None for non-reflexive, non-transitive, non-symmetric


        """
        if obo_name not in self._instances:
            self.obo_name = obo_name
            self.symmetry = symmetry
            self.transitivity = reflexivity
            self.reflexivity = reflexivity
            self.complementary = complementary or ''
            self.prefix = prefix or ''
            self.direction = direction or ''
            self.comment = comment or ''
            self.aliases = aliases or list()
            self._instances[obo_name] = self
            for alias in self.aliases:
                self._instances[alias] = self

    def complement(self):
        """Return the complementary relationship of self.

        Raises:
            ValueError: if the relationship has a complementary
                        which was not defined.

        Returns:
            complementary (Relationship): the complementary relationship.

        Example:

            >>> from pronto.relationship import Relationship
            >>> print(Relationship('has_part').complement())
            Relationship(part_of)
            >>> print(Relationship('has_units').complement())
            None

        """

        if self.complementary:

            #if self.complementary in self._instances.keys():
            try:
                return self._instances[self.complementary]
            except KeyError:
                raise ValueError('{} has a complementary but it was not defined !')

        else:
            return None

    def __repr__(self):
        """Overloaded :obj:`object.__repr__`"""
        return "Relationship({})".format(self.obo_name)

#    @classmethod
#    @staticmethod
    def __new__(cls, obo_name, *args, **kwargs):
        """Overloaded :obj:`object.__new__` method that __memoizes__ the objects.

        This allows to do the following (which is frecking cool):

            >>> Relationship('has_part').direction
            'topdown'

        The Python syntax is overloaded, and what looks like a object
        initialization in fact retrieves an existing object with all its
        properties already set ! The Relationship class behaves like a
        factory of its own objects !

        Todo:
            * Add a warning for unknown relationship (the goal being to
              instantiate every known ontology relationship and even
              allow instatiation of file-defined relationships).

        """
        if obo_name in cls._instances:
            return cls._instances[obo_name]
        else:
            return super(Relationship, cls).__new__(cls)

    #@pronto.utils.classproperty
    @classmethod
    def topdown(cls):
        """Get all topdown Relationship instances

        Returns:
            :obj:`generator`

        Example:

            >>> from pronto import Relationship
            >>> for r in Relationship.topdown():
            ...    print(r)
            Relationship(can_be)
            Relationship(has_part)

        """
        return tuple(pronto.utils.unique_everseen(r for r in cls._instances.values() if r.direction=='topdown'))

    #@pronto.utils.classproperty
    @classmethod
    def bottomup(cls):
        """Get all bottomup Relationship instances

        Example:

            >>> from pronto import Relationship
            >>> for r in Relationship.bottomup():
            ...    print(r)
            Relationship(is_a)
            Relationship(part_of)

        """
        return tuple(pronto.utils.unique_everseen(r for r in cls._instances.values() if r.direction=='bottomup'))

    @pronto.utils.classproperty
    def lock(self):
        """A multiprocessing.Lock provided at a class level

        This allows to use pronto's Relationship objects in a multiprocessed
        environment.

        Todo:
            * Use the lock within Relationship.__init__ method and Relationship.topdown
              and Relationship.bottomup
            * Write a classmethod Relationship.instances that wraps the process of acquiring
              the lock during iteration to avoid data races within iteration if other Relationship
              if created (! may end in a deadlock)

        """
        return self._lock

    @pronto.utils.classproperty
    def tlock(self):
        """A threading.Lock provided at a class level
        """
        return self._tlock

    def __getstate__(self):
        pass

    def __getnewargs__(self):
        return (self.obo_name,)

    def __setstate__(self, *args, **kwargs):
        pass


Relationship('is_a', symmetry=False, transitivity=True,
                    reflexivity=True, complementary='can_be',
                    direction='bottomup')

Relationship('can_be', symmetry=False, transitivity=True,
                    reflexivity=True, complementary='is_a',
                    direction='topdown')

Relationship('has_part', symmetry=False, transitivity=True,
                        reflexivity=True, complementary='part_of',
                        direction='topdown')

Relationship('part_of', symmetry=False, transitivity=True,
                        reflexivity=True, complementary='has_part',
                        direction='bottomup', aliases=['is_part'])

Relationship('has_units', symmetry=False, transitivity=False,
                          reflexivity=None)

Relationship('has_domain', symmetry=False, transitivity=False,
                           reflexivity=None)



# coding: utf-8
"""
pronto.utils
============

This module contains some functions that are used in different parts
of the pronto library, as well as the definition of ProntoWarning.

Todo:
    * Maybe add a ProntoError class ?
"""





#import functools
#import errno
#import os
#import signal
import itertools
import atexit

import multiprocessing
import multiprocessing.pool
import multiprocessing.queues



try:
    from urllib.error import URLError, HTTPError
except ImportError:
    from urllib2 import URLError, HTTPError
    itertools.filterfalse = itertools.ifilterfalse

try:
    from lxml.etree import XMLSyntaxError as ParseError
except ImportError:
    from xml.etree.ElementTree import ParseError




class TimeoutError(IOError):
    pass



#def memoize(obj):
#    cache = obj.cache = {}
#
#    @functools.wraps(obj)
#    def memoizer(*args, **kwargs):
#        if args not in cache:
#            cache[args] = obj(*args, **kwargs)
#        return cache[args]
#
#    return memoizer


# def timeout(seconds=60, error_message=os.strerror(errno.ETIME)):

#     def decorator(func):

#         def _handle_timeout(signum, frame):
#             raise TimeoutError(error_message)

#         @functools.wraps(func)
#         def wrapper(*args, **kwargs):

#             try:
#                 signal.signal(signal.SIGALRM, _handle_timeout)
#                 signal.alarm(seconds)
#             except ValueError:
#                 pass
#             try:
#                 result = func(*args, **kwargs)
#             finally:
#                 signal.alarm(0)
#             return result
#         return wrapper

#     return decorator

class classproperty(object):
    """
    A decorator that allows to set properties on class level.

    """

    def __init__(self, fget):
        self.__doc__ = fget.__doc__
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class ProntoWarning(Warning):
    """A warning raised by pronto.

    Example:
        >>> from pronto import Ontology
        >>> import warnings
        >>> with warnings.catch_warnings(record=True) as w:
        ...    # the following ontology always has import issues (no URI in imports)
        ...    ims = Ontology('https://raw.githubusercontent.com/beny/imzml'
        ...                   '/master/data/imagingMS.obo')
        >>> print(w[-1].category)
        <class 'pronto.utils.ProntoWarning'>

    """

    def __init__(self, *args, **kwargs):

        super(ProntoWarning, self).__init__(*args, **kwargs)
        #self.__suppress_context__ = True


class _NoDaemonProcess(multiprocessing.Process): # pragma: no cover
    # make 'daemon' attribute always return False

    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class ProntoPool(multiprocessing.pool.Pool): # pragma: no cover
    """A non-daemonized pool provided for convenience.

    Allows to perform ontology parsing through a pool of non-daemonized
    workers while inheriting all methods and attributes of multiprocessing.pool.Pool.

    Example:

        >>> from pronto import Ontology
        >>> from pronto.utils import ProntoPool
        >>> enm = [ #ontologies from the eNanoMapper project
        ... "http://purl.enanomapper.net/onto/external/chebi-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/bao-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/bfo-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/ccont-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/cheminf-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/chmo-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/efo-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/envo-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/go-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/hupson-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/iao-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/ncit-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/npo-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/oae-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/obcs-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/obi-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/pato-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/sio-slim.owl",
        ... "http://purl.enanomapper.org/onto/external/uo-slim.owl"]
        >>> pool = ProntoPool()

        >>> from functools import partial
        >>> enm_onto = pool.map(partial(Ontology, timeout=None), enm)
        >>> pool.close()

    """
    _instances = []
    Process = _NoDaemonProcess

    def __init__(self, *args, **kwargs):
        super(ProntoPool, self).__init__(*args, **kwargs)
        self._instances.append(self)

    @classmethod
    def _close_all(cls):
        for pool in cls._instances:
            pool.close()

        for pool in cls._instances:
            pool.join()
atexit.register(ProntoPool._close_all)


class SharedCounter(object):
    """ A synchronized shared counter.

    The locking done by multiprocessing.Value ensures that only a single
    process or thread may read or write the in-memory ctypes object. However,
    in order to do n += 1, Python performs a read followed by a write, so a
    second process may read the old value before the new one is written by
    the first process. The solution is to use a multiprocessing.Lock to
    guarantee the atomicity of the modifications to Value.

    This class comes almost entirely from Eli Bendersky's blog:
    http://eli.thegreenplace.net/2012/01/04/shared-counter-with-
        pythons-multiprocessing/

    --------------------------------------------------------------------------

    Solution of issue:
        https://github.com/vterron/lemon/issues/11
    implemented in:
        https://github.com/vterron/lemon/commit/9ca6b4b1212228dbd4f69b88aaf88b12952d7d6f

    """

    def __init__(self, n = 0):
        self.count = multiprocessing.Value('i', n)

    def increment(self, n = 1):
        """ Increment the counter by n (default = 1) """
        with self.count.get_lock():
            self.count.value += n

    @property
    def value(self):
        """ Return the value of the counter """
        return self.count.value


class Queue(multiprocessing.queues.Queue):
    """ A portable implementation of multiprocessing.Queue.

    Because of multithreading / multiprocessing semantics, Queue.qsize() may
    raise the NotImplementedError exception on Unix platforms like Mac OS X
    where sem_getvalue() is not implemented. This subclass addresses this
    problem by using a synchronized shared counter (initialized to zero) and
    increasing / decreasing its value every time the put() and get() methods
    are called, respectively. This not only prevents NotImplementedError from
    being raised, but also allows us to implement a reliable version of both
    qsize() and empty().

    --------------------------------------------------------------------------

    Solution of issue:
        https://github.com/vterron/lemon/issues/11
    implemented in:
        https://github.com/vterron/lemon/commit/9ca6b4b1212228dbd4f69b88aaf88b12952d7d6f

    """

    def __init__(self, *args, **kwargs):
        try:
            super(Queue, self).__init__(*args,**kwargs)
        except TypeError:
            super(Queue, self).__init__(*args, ctx=multiprocessing.get_context(), **kwargs)
        self.size =  SharedCounter(0)

    def put(self, *args, **kwargs):
        super(Queue, self).put(*args, **kwargs)
        self.size.increment(1)

    def get(self, *args, **kwargs):
        x = super(Queue, self).get(*args, **kwargs)
        self.size.increment(-1)
        return x

    def qsize(self):
        """ Reliable implementation of multiprocessing.Queue.qsize() """
        return self.size.value

    def empty(self):
        """ Reliable implementation of multiprocessing.Queue.empty() """
        return not self.qsize() > 0




def _ontologize(x):
    Ontology, path, import_depth = x
    try:
        return Ontology(path, import_depth=import_depth-1)
    except (IOError, OSError, URLError, HTTPError, ParseError) as e:
        return ("{} occured during import of {}".format(type(e).__name__, path),
                ProntoWarning)

def explicit_namespace(attr, nsmap):
    """Explicits the namespace in an attribute name.

    Parameters
        attr (str): the attribute with its abbreviated namespace
        nsmap (dict): the namespace map

    Example:
        >>> from pronto.utils import explicit_namespace
        >>> ns = {'owl':'http://www.w3.org/2002/07/owl#'}
        >>> explicit_namespace('owl:Class', ns)
        '{http://www.w3.org/2002/07/owl#}Class'

    """
    prefix, term = attr.split(':', 1)
    return "".join(['{', nsmap[prefix], '}', term])

def format_accession(accession, nsmap=None):
    """Formats an accession URI/string to the YY:XXXXXXX token format.

    Parameters:
        accession (str): the URI to be formatted
        nsmap (dict): namespaces that can be found at the beginning of the
            accession that we want to get rid of.

    Example:
        >>> from pronto.utils import format_accession
        >>> format_accession('UO_1000003')
        'UO:1000003'
        >>> ns = {'obo':'http://purl.obolibrary.org/obo/'}
        >>> format_accession('http://purl.obolibrary.org/obo/IAO_0000601', ns)
        'IAO:0000601'

    """

    if nsmap is not None:
        for v in nsmap.values():
            accession = accession.replace(v, '')

    if not accession.startswith('_'):
        accession = accession.replace('_', ':')

    return accession

def unique_everseen(iterable):
    """List unique elements, preserving order. Remember all elements ever seen."""
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen = set()
    seen_add = seen.add

    for element in itertools.filterfalse(seen.__contains__, iterable):
        seen_add(element)
        yield element




"""
pronto.parser

This module contains all the parsers used to parse an ontology. The base class
Parser can be derived to provide additional parsers to ontology files.
"""

import multiprocessing
import six

import pronto.term
import pronto.relationship
from pronto.parser import Parser



class _OboClassifier(multiprocessing.Process): # pragma: no cover

    def __init__(self, queue, results, *args, **kwargs):

        super(_OboClassifier, self).__init__()

        self.queue = queue
        self.results = results

    def run(self):



        while True:


            term = self.queue.get()

            if term is None:
                #self.queue.task_done()
                break


            self.results.put(self._classify(term))

        #return {tid: pronto.term.Term(tid, name, desc, relations, term)}

    @staticmethod
    def _classify(term):

        relations = {}

        if 'relationship' in term:
            #pronto.relationship.Relationship.lock.acquire()
            #try:
            _OboClassifier._process_relationship(term)
            #finally:
            #    pronto.relationship.Relationship.lock.release()

        #if 'is_a' in term:
        try:
            term[pronto.relationship.Relationship('is_a')] = [ term['is_a'].format() ]
            del term['is_a']
        except AttributeError:
            term[pronto.relationship.Relationship('is_a')] = term['is_a']
            del term['is_a']
        except KeyError:
            pass
            #term[pronto.relationship.Relationship('is_a')] = term['is_a'] if isinstance(term['is_a'])
            #del term['is_a']

        for key in tuple(term.keys()):
            if isinstance(key, pronto.relationship.Relationship):
                relations = _OboClassifier._extract_relationship(term, key, relations)

        desc = _OboClassifier._extract_description(term)
        tid, name = _OboClassifier._extract_name_and_id(term)

        return (tid, name, desc, relations, term)

    @staticmethod
    def _process_relationship(term):

        # If term is a str
        try:

            name, value = term['relationship'].split(' ', 1)

            name = pronto.relationship.Relationship(name)

            try:
                term[name].append(value)
            except KeyError:
                term[name] = [value]

        # If term is a list
        except AttributeError:

            for r in term['relationship']:
                name, value = r.split(' ', 1)

                name = pronto.relationship.Relationship(name)

                try:
                    term[name].append(value)
                except KeyError:
                    term[name] = [value]

        del term['relationship']

    @staticmethod
    def _extract_relationship(term, rship, relations):

        #if rship.obo_name in term.keys():
        #if isinstance(term[rship], list):
        try:
            relations[rship.obo_name] = [ x.split(' !')[0] for x in term[rship] ]
        except AttributeError:
            relations[rship.obo_name] = [ term[rship].split(' !')[0]]



        del term[rship]

        return relations

    @staticmethod
    def _extract_description(term):
        try:
        #if 'def' in term.keys():
            desc = term['def']
            del term['def']
            return desc
        except KeyError:
            return ''

    @staticmethod
    def _extract_name_and_id(term):
        if 'id' not in term.keys():
            return '', ''
        if 'name' in term.keys():
            tid, name = term['id'], term['name']
            del term['name']
            del term['id']
            return tid, name
        else:
            tid = term['id'][0]
            name = ''
            del term['id']
            return tid, name




class OboParser(Parser):
    """A parser for the Obo format.
    """

    def __init__(self):
        super(OboParser, self).__init__()
        self.extension = '.obo'

        self._tempterm = {}
        self._typedef = []
        self._meta = {}

    def hook(self, *args, **kwargs):
        """Returns True if the file is an Obo file (extension is .obo)"""

        if 'path' in kwargs:
            return kwargs['path'].endswith(self.extension)

    def read(self, stream):
        """Read the stream and extract information in a 'raw' format."""
        #self._rawterms, = multiprocessing.JoinableQueue()
        self._typedef, self._meta = [], {}

        self.init_workers(_OboClassifier)


        IN = 'meta'
        for line in stream:
            try:
                line = line.strip().decode('utf-8')
            except AttributeError:
                line = line.strip()

            IN = self._check_section(line) or IN
            if ': ' in line:
                self._parse_line_statement(line, IN)

        self._rawterms.put(self._tempterm)


    def makeTree(self):
        """Create the proper ontology Tree from raw terms"""

        while not self._terms.empty() or not self._rawterms.empty(): #self._terms.qsize() > 0 or self._rawterms.qsize() > 0:
            d = self._terms.get()
            self.terms[d[0]] = pronto.term.Term(
                d[0], d[1], d[2], {pronto.relationship.Relationship(k):v for k,v in six.iteritems(d[3])}, d[4]
            )

        self.shut_workers()

    def metanalyze(self):
        """Analyze metadatas extracted from the beginning of the file."""

        for key,value in six.iteritems(self._meta):
            if key != 'remark':
                self._parse_statement(key, value)
            else:
                for remark in value:
                    self._parse_remark(remark)

    def manage_imports(self):
        """Get metadatas concerning imports."""
        try:
        #if 'import' in self.meta: #.keys():
            self.imports = set(self.meta['import'])

        except KeyError:
            pass #self.imports = set()

    def _parse_remark(self, remark):
        """Parse a remark and add results to self.meta."""
        try:
            key_remark, value_remark = remark.split(': ', 1)
            try:
                self.meta[key_remark].append(value_remark)
            except KeyError:
                self.meta[key_remark] = [value_remark]

        except ValueError:
            try:
                self.meta['remark'].append(remark)
            except KeyError:
                self.meta['remark'] = [remark]

    def _parse_statement(self, key, value):
        """Parse a ``key: value`` statement in the ontology file."""

        try:
            self.meta[key].extend(value)
        except TypeError:
            self.meta[key].append(value)
        except KeyError:
            try:
                self.meta[key] = [ value.format() ]
            except AttributeError:
                self.meta[key] = value

    def _get_dict_to_update(self, IN):
        """Returns the right dictionnary to use"""
        if IN=='meta':
            return self._meta
        elif IN=='term':
            return self._tempterm
        elif IN=='typedef':
            return self._typedef[-1]

    def _check_section(self, line):
        """Check if current section must be changed."""
        if '[Term]' in line:
            if self._tempterm:
                self._rawterms.put(self._tempterm)
            self._tempterm = {}
            return 'term'

        elif '[Typedef]' in line:
            self._typedef.append({})
            return 'typedef'

    def _parse_line_statement(self, line, IN):
        k, v = line.split(': ', 1)
        to_update = self._get_dict_to_update(IN)

        try:
            to_update[k].append(v)
        except AttributeError:
            to_update[k] = [to_update[k], v]
        except KeyError:
            to_update[k] = v






OboParser()




import functools
import os
import multiprocessing
import six
""
try:
    import lxml.etree as etree
    from lxml.etree import XMLSyntaxError as ParseError
except ImportError: # pragma: no cover
    try:
        import xml.etree.cElementTree as etree
        from xml.etree.cElementTree import ParseError
    except ImportError:
        import xml.etree.ElementTree as etree
        from xml.etree.ElementTree import ParseError


from pronto.parser import Parser
from pronto.relationship import Relationship
import pronto.utils

class _OwlXMLClassifier(multiprocessing.Process): # pragma: no cover

    def __init__(self, queue, results):

        super(_OwlXMLClassifier, self).__init__()

        self.queue = queue
        self.results = results
        nsmap = {'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                 'rdfs': 'http://www.w3.org/2000/01/rdf-schema#'}

        self.nspaced = functools.partial(pronto.utils.explicit_namespace, nsmap=nsmap)
        self.accession = functools.partial(pronto.utils.format_accession, nsmap=nsmap)

    def run(self):

        while True:


            term = self.queue.get()


            if term is None:
                break

            try:
                classified_term = self._classify(etree.fromstring(term))
            except ParseError:
                self.results.put((None, None))
                break


            if classified_term:
                self.results.put(classified_term)

    def _classify(self, term):
        """
        Map raw information extracted from each owl Class.

        The raw data (in an etree.Element object) is extracted to a proper
        dictionnary containing a Term referenced by its id, which is then
        used to update :attribute:terms

        Todo:
            * Split into smaller methods to lower code complexity.
        """

        if not term.attrib:
           return {}

        tid = self.accession(term.get(self.nspaced('rdf:about')))

        term_dict = {'name':'', 'relations': {}, 'desc': ''}

        translator = [
            {'hook': lambda c: c.tag == self.nspaced('rdfs:label'),
             'callback': lambda c: c.text,
             'dest': 'name',
             'action': 'store'
            },
            {
             'hook': lambda c: (c.tag == self.nspaced('rdfs:subClassOf')) \
                               and (self.nspaced('rdf:resource') in c.attrib),
             'callback': lambda c: self.accession(c.get(self.nspaced('rdf:resource')) or c.get(self.nspaced('rdf:about'))),
             'dest': 'relations',
             'action': 'list',
             'list_to': 'is_a',
            },
            {'hook': lambda c: c.tag == self.nspaced('rdfs:comment'),
             'callback': lambda c: self.parse_comment(c.text),
             'action': 'update'
            }
        ]

        for child in term.iter():#term.iterchildren():

            for rule in translator:

                if rule['hook'](child):

                    if rule['action'] == 'store':
                        term_dict[rule['dest']] = rule['callback'](child)

                    elif rule['action'] == 'list':

                        try:
                            term_dict[rule['dest']][rule['list_to']].append(rule['callback'](child))
                        except KeyError:
                            term_dict[rule['dest']][rule['list_to']] = [rule['callback'](child)]


                    elif rule['action'] == 'update':
                        term_dict.update(rule['callback'](child))


                    #break

        #if ':' in tid: #remove administrative classes
        return (tid, term_dict)#{tid: pronto.term.Term(tid, **term_dict)}
        #else:
        #    return {}

    @staticmethod
    def parse_comment(comment):
        """Parse an rdfs:comment to extract information.

        Owl contains comment which can contain additional metadata (specially
        when the Owl file was converted from Obo to Owl). This function parses
        the comment to try to extract those metadata.

        Parameters:
            comment (str): if containing different sections (such as 'def:',
                'functional form' or 'altdef:'), the value of those sections will
                be returned in a dictionnary. If there are not sections, the
                comment is interpreted as a description

        Todo:
            * Add more parsing special cases

        """
        if comment is None:
            return {}

        commentlines = comment.split('\n')
        parsed = {}

        for (index, line) in enumerate(commentlines):

            line = line.strip()

            if line.startswith('Functional form:'):
                #if not 'other' in parsed.keys():
                #    parsed['other'] = dict()

                try:
                    parsed['other']['functional form'] = "\n".join(commentlines[index:])
                except KeyError:
                    parsed['other'] = {'functional form': "\n".join(commentlines[index:])}

                break

            if line.startswith('def:'):
                parsed['desc'] = line.split('def:')[-1].strip()

            elif ': ' in line:
                ref, value = [x.strip() for x in line.split(': ', 1)]

                try:
                    parsed['other'][ref].append(value)
                except KeyError:
                    try:
                        parsed['other'][ref] = [value]
                    except KeyError:
                        parsed['other'] = {ref: [value] }


                #if not 'other' in parsed.keys():
                #    parsed['other'] = {}
                #if not ref in parsed['other']:
                #    parsed['other'][ref] = []
                #parsed['other'][ref].append(value)

            else:
                if not 'desc' in parsed:
                    parsed['desc'] = "\n".join(commentlines[index:])
                    break

        if not 'desc' in parsed and 'other' in parsed:
            #if 'tempdef' in parsed['other'].keys():
            try:
                parsed['desc'] = parsed['other']['tempdef']
                del parsed['other']['tempdef']
            except KeyError:
                pass

            #if 'altdef' in parsed['other'].keys():
            try:
                parsed['desc'] = parsed['other']['altdef']
                del parsed['other']['altdef']
            except KeyError:
                pass


        return parsed

class OwlXMLParser(Parser):
    """A parser for the owl xml format.
    """

    def __init__(self, daemon=False):
        super(OwlXMLParser, self).__init__()
        self._tree = None
        self._ns = {}
        self.extensions = ('.owl', '.xml', '.ont')

    def hook(self, *args, **kwargs):
        """Returns True if the file is an Owl file (extension is .owl)"""
        if 'path' in kwargs:
            return os.path.splitext(kwargs['path'])[1] in self.extensions

    def read(self, stream):
        """
        Parse the content of the stream
        """

        self.init_workers(_OwlXMLClassifier)

        events = ("end", "start-ns")

        owl_imports, owl_class, rdf_resource = "", "", ""

        context = etree.iterparse(stream, events=events)

        for event, element in context: #etree.iterparse(stream, #huge_tree=True,
                                       #                events=events):

            if element is None:
                break

            if event == "start-ns":
                self._ns.update({element[0]:element[1]})

                if element[0]== 'owl':
                    owl_imports = "".join(["{", element[1], "}", "imports"])
                    owl_class = "".join(["{", element[1], "}", "Class"])
                elif element[0] == 'rdf':
                    rdf_resource = "".join(["{", element[1], "}", "resource"])

                del event
                del element
                continue

            elif element.tag==owl_imports:
                self.imports.add(element.attrib[rdf_resource])


            elif element.tag==owl_class:
                if element.attrib:
                    self._rawterms.put(etree.tostring(element))

            else:
                continue

            element.clear()

            try:
                for ancestor in element.xpath('ancestor-or-self::*'):
                    while ancestor.getprevious() is not None:
                        del ancestor.getparent()[0]
            except AttributeError:
                pass

            del element

        del context

    def makeTree(self):
        """
        Maps :function:_classify to each term of the file via a ThreadPool.

        Once all the raw terms are all classified, the :attrib:terms dictionnary
        gets updated.

        Arguments:
            pool (Pool): a pool of workers that is used to map the _classify
                function on the terms.
        """
        #terms_elements = self._tree.iterfind('./owl:Class', self._ns)
        #for t in pool.map(self._classify, self._elements):
        #    self.terms.update(t)

        accession = functools.partial(pronto.utils.format_accession, nsmap=self._ns)

        while not self._terms.empty() or not self._rawterms.empty(): #self._terms.qsize() > 0 or self._rawterms.qsize() > 0:

            tid, d = self._terms.get()

            if tid is None and d is None:
                break

            tid = pronto.utils.format_accession(tid, self._ns)

            d['relations'] = { Relationship(k):[accession(x) for x in v] for k,v in six.iteritems(d['relations']) }

            self.terms[tid] = pronto.term.Term(tid, **d)


        # if 'tid' in locals() and locals()['tid'] is None:
        #     for p in self._processes:
        #         self._rawterms.put(None)
        #     self.shut_workers()
        #     #return ParseError
        # else:
        self.shut_workers()


    def manage_imports(self):
        pass
        #nspaced = functools.partial(pronto.utils.explicit_namespace, nsmap=self._ns)
        #for imp in self._tree.iterfind('./owl:Ontology/owl:imports', self._ns):
        #    path = imp.attrib[nspaced('rdf:resource')]
        #    if path.endswith('.owl'):
        #        self.imports.append(path)

    def metanalyze(self):
        """
        Extract metadata from the headers of the owl file.

        Todo:
            * Implement that method !
        """
        pass


OwlXMLParser()



import warnings
import weakref
import multiprocessing
#import atexit

import pronto.utils

__all__ = ["Parser", "OboParser", "OwlXMLParser"]



class Parser(object):
    """An abstract parser object.
    """

    _instances = {}

    def __init__(self, timeout=None):
        self.terms = {}
        self.meta = {}
        self.imports = []
        self._instances[type(self).__name__] = self
        self._rawterms, self._terms, self._processes= None, None, None

    #@pronto.utils.timeout(0)
    def parse(self, stream):
        """
        Parse the ontology file.

        Parameters
            stream (io.StringIO): A stream of the ontology file.
        """

        self.terms, self.meta, self.imports = {}, {}, set()

        self.read(stream)
        self.makeTree()
        self.metanalyze()
        self.manage_imports()

        return self.meta, self.terms, self.imports

    def hook(self, *args, **kwargs):
        """Defines when the parser is used

        For Obo and Owl, based on file extension (altough that may need to
        change depending on Owl format).
        """
        raise NotImplementedError

    def read(self, stream):
        """Reads and preprocesses the stream.
        """
        raise NotImplementedError

    def makeTree(self):
        raise NotImplementedError

    def metanalyze(self):
        raise NotImplementedError

    def manage_imports(self):
        raise NotImplementedError

    def init_workers(self, ParserProcess, *args, **kwargs):
        self._rawterms = pronto.utils.Queue()#multiprocessing.Queue()
        self._terms = pronto.utils.Queue()#multiprocessing.Queue()
        self._processes = []

        for _ in range(multiprocessing.cpu_count() * 2):
            self._processes.append(ParserProcess(self._rawterms, self._terms))

        for p in self._processes:
            p.start()

    def shut_workers(self):
        self._rawterms.close()
        self._terms.close()

        #while self._processes:
        for p in multiprocessing.active_children():
            #p = self._processes.pop()
            p.terminate()

        #for p in self._processes:
        #    p.join()


    @classmethod
    def __del__(cls):

        try:
            for p in cls.processes:
                p.terminate()

            cls._rawterms.close()
            cls._terms.close()
            del cls._terms
            del cls._rawterms
            del cls.processes

        except AttributeError:
            pass



#atexit.register(Parser.__del__)


from pronto.parser.obo import OboParser

try:
   from pronto.parser.owl import OwlXMLParser
except ImportError:
   warnings.warn("You don't seem to have lxml installed on your machine, "
                 ".owl parsing will be disabled", pronto.utils.ProntoWarning)



# coding: utf-8
"""
pronto.term
===========

This module defines the classes Term and TermList.
"""

import six

from pronto.relationship import Relationship


class Term(object):
    """ An ontology term.

    Examples:

        >>> from pronto import *
        >>> new_term = Term('TR:001', 'new term', 'a new term')
        >>> linked_term = Term('TR:002', 'other new', 'another term',
        ...                    { Relationship('is_a'): 'TR:001'})

        >>> ms = Ontology('https://raw.githubusercontent.com/'
        ...               'HUPO-PSI/psi-ms-CV/master/psi-ms.obo')
        >>> type(ms['MS:1000015'])
        <class 'pronto.term.Term'>


    """

    def __init__(self, tid, name, desc='', relations=None, other=None):
        """

        Parameters:
            tid (str): the Term id (e.g. MS:1000031)
            name (str): the name of the Term in human language
            desc (str): a description of the Term
            relations (dict): a dictionary containing the other
                terms the Term is in a relationship with.
            other (dict): other information about the term

        """
        self.id = tid
        self.name = name
        self.desc = desc
        self.relations = relations or {}
        self.other = other or {}
        self._rchildren  = {}
        self._rparents = {}
        self._children = None
        self._parents = None

    def __repr__(self):
        return "<{}: {}>".format(self.id, self.name)

    @property
    def parents(self):
        """The parents of the Term.

        Returns:
            :obj:`pronto.TermList`:
            a TermList containing all parents of the Term
            (other terms with which this Term has a "bottomup"
            relationship)

        Example:

            >>> for p in ms['MS:1000532'].parents: print(p.desc)
            "Thermo Finnigan software for data acquisition and analysis." [PSI:MS]
            "Acquisition software." [PSI:MS]
            "Analysis software." [PSI:MS]
            "Data processing software." [PSI:MS]

        """

        if self._parents is None:
            bottomups = tuple(Relationship.bottomup())

            self._parents = TermList()
            self._parents.extend(
                [ other
                    for rship,others in six.iteritems(self.relations)
                        for other in others
                            if rship in bottomups
                ]

            )

        return self._parents

    @property
    def children(self):
        """The children of the Term.

        Returns:
            :obj:`pronto.TermList`:

            a TermList containing all parents of the Term
            (other terms with which this Term has a "topdown"
            relationship)

        Example:

            >>> ms['MS:1000452'].children
            [<MS:1000530: file format conversion>, <MS:1000543: data processing action>]

        """

        if self._children is None:

            topdowns = tuple(Relationship.topdown())
            self._children = TermList()
            self._children.extend(
                [ other
                    for rship,others in six.iteritems(self.relations)
                        for other in others
                            if rship in topdowns
                ]

            )

        return self._children



            #for rship in Relationship.topdown():
            #    if rship in self.relations:
            #        self._children.extend(self.relations[rship])
        #return self._children

    @property
    def obo(self):
        """Displays the term as an obo Term entry

        Example:

            >>> print(ms['MS:1000031'].obo)
            [Term]
            id: MS:1000031
            name: instrument model
            def: "Instrument model name not including the vendor's name." [PSI:MS]
            relationship: part_of: MS:1000463 ! instrument

        """



        obo =  "".join([ '[Term]', '\n',
        #obo +=
                         'id: ', self.id, '\n',

        #obo +=
                         'name: ', self.name if self.name is not None else '', '\n'])

        if self.desc:
            obo = "".join([obo, 'def: ', self.desc, '\n'])

        # add more bits of information
        for k,v in six.iteritems(self.other):
            if isinstance(v, list):
                for x in v:
                    obo = "".join([obo, k, ': ', x, '\n'])
            else:
                obo = "".join([obo,k, ': ', v, '\n'])

        # add relationships (only bottom up ones)

        for relation in Relationship.bottomup():
            try:
                for companion in self.relations[relation]:

                    if relation is not Relationship('is_a'):
                        obo = "".join([obo, 'relationship: '])
                    obo = "".join([obo, relation.obo_name, ': '])

                    try:
                        obo = "".join([obo, companion.id, ' ! ', companion.name, '\n'])
                    except AttributeError:
                        obo = "".join([obo,companion, '\n'])

            except KeyError:
                continue

        return obo.rstrip()

    @property
    def __deref__(self):
        """A dereferenced relations dictionary only contains other Terms id
        to avoid circular references when creating a json.
        """
        return {
            'id': self.id,
            'name': self.name,
            'other': self.other,
            'desc': self.desc,
            'relations': {k.obo_name:v.id for k,v in six.iteritems(self.relations)}
         }

    def __getstate__(self):

        return (
            self.id,
            self.name,
            tuple((k,v) for k,v in six.iteritems(self.other)),
            self.desc,
            tuple((k.obo_name,v.id) for k,v in six.iteritems(self.relations)),
        )

    def __setstate__(self, state):

        self.id = state[0]
        self.name = state[1]
        self.other = {k:v for (k,v) in state[2]}
        self.desc = state[3]
        self.relations = {Relationship(k):v for k,v in state[4]}

    def _empty_cache(self):
        """
        Empties the cache of the Term's memoized functions.
        """
        self._children, self._parents = None, None
        self._rchildren, self._rparents = {}, {}

    def rchildren(self, level=-1, intermediate=True):
        """Create a recursive list of children.

        Note that the :param:`intermediate` can be used to include every
        child to the returned list, not only the most nested ones.

        Parameters:
            level (int): The depth level to continue fetching children from
                (default is -1, to get children to the utter depths)
            intermediate (bool): Also include the intermediate children
                (default is True)

        Returns:
            :obj:`pronto.TermList`:
            The recursive children of the Term following the parameters

        """
        try:
            return self._rchildren[(level, intermediate)]

        except KeyError:

            rchildren = []

            if self.children and level:

                if intermediate or level==1:
                    rchildren.extend(self.children)

                for child in self.children:
                    rchildren.extend(child.rchildren(level=level-1,
                                                     intermediate=intermediate))

            rchildren = TermList(set(rchildren))
            self._rchildren[(level, intermediate)] = rchildren
            return rchildren

    def rparents(self, level=-1, intermediate=True):
        """Create a recursive list of children.

        Note that the :param:`intermediate` can be used to include every
        parents to the returned list, not only the most nested ones.

        Parameters:
            level (int): The depth level to continue fetching parents from
                (default is -1, to get parents to the utter depths)
            intermediate (bool): Also include the intermediate parents
                (default is True)

        Returns:
            :obj:`pronto.TermList`:
            The recursive children of the Term following the parameters

        """
        try:
            return self._rparents[(level, intermediate)]

        except KeyError:

            rparents = []

            if self.parents and level:

                if intermediate or level==1:
                    rparents.extend(self.parents)

                for parent in self.parents:
                    rparents.extend(parent.rparents(level=level-1,
                                                     intermediate=intermediate))

            rparents = TermList(set(rparents))
            self._rparents[(level, intermediate)] = rparents
            return rparents


class TermList(list):
    """A list of Terms.

    TermList behaves exactly like a list, except it contains shortcuts to
    generate lists of terms' attributes.

    Example:
        >>> from pronto import *
        >>> nmr = Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl')
        >>> type(nmr['NMR:1000031'].children)
        <class 'pronto.term.TermList'>

        >>> nmr['NMR:1000031'].children.id
        ['NMR:1000122', 'NMR:1000156', 'NMR:1000157', 'NMR:1000489']
        >>> nmr['NMR:1400014'].relations[Relationship('is_a')]
        [<NMR:1400011: cardinal part of NMR instrument>]


    .. tip::
        It is also possible to call Term methods on a TermList to
        create a set of terms::

            >>> nmr['NMR:1000031'].rchildren(3, False).rparents(3, False).id
            ['NMR:1000031']

    """

    def __init__(self, *elements):
        """
        """
        list.__init__(self, *elements)
        self._check_content()

    def _check_content(self):
        try:
            [ term.id for term in self ]
        except AttributeError:
            raise TypeError('TermList can only contain Terms.')

        #for term in self:
        #    if not isinstance(term, Term):
        #        raise TypeError('TermList can only contain Terms.')

    # def __getattr__(self, attr, *args, **kwargs):
    #     if attr in ('children', 'parents'):
    #         return TermList( [ y for x in self for y in getattr(x, attr)] )
    #     elif attr in ('rparents', 'rchildren'):
    #         #: we create a new method to allow the user
    #         #: to use, for instance, ``x.rchildren(3).rparents(2)``
    #         #: (this actually behaves as if you mapped the method
    #         #: on all terms of the TermList)

    #         #def mapped(level=-1, intermediate=True):
    #         #    t = TermList(set([ y for x in self
    #         #            for y in getattr(x, attr)(level, intermediate) ]))
    #         #    return t
    #         #return mapped

    #         return self.__dict__[attr]

    #     elif attr in ('id', 'name', 'desc', 'other', 'obo'):
    #         return [getattr(x, attr) for x in self]
    #     else:
    #         getattr(list, attr)

    def rparents(self, level=-1, intermediate=True):
        return TermList(set(
            [y for x in self for y in x.rparents(level, intermediate)]
        ))

    def rchildren(self, level=-1, intermediate=True):
        return TermList(set(
            [y for x in self for y in x.rchildren(level, intermediate)]
        ))

    @property
    def children(self):
        return TermList( [ y for x in self for y in x.children] )

    @property
    def parents(self):
        return TermList( [ y for x in self for y in x.parents] )

    #elif attr in ('id', 'name', 'desc', 'other', 'obo'):
    #         return [getattr(x, attr) for x in self]

    @property
    def id(self):
        return [x.id for x in self]

    @property
    def name(self):
        return [x.name for x in self]

    @property
    def desc(self):
        return [x.desc for x in self]

    @property
    def other(self):
        return [x.other for x in self]

    @property
    def obo(self):
        return [x.obo for x in self]

    def __getstate__(self):
        return (x for x in self)

    def __setstate__(self, state):
        self.extend(state)

    def __contains__(self, term):
        """
        Todo:
            write doc & test
        """
        return term in self.id or any(t for t in self if t==term)


